/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */


#include <string.h>
#include "mygl.h"
#ifndef VISUAL_CPP
#include <new.h>
#else
#define WANT_NOTHROW
#include "porting.h"
#endif
#include "Textures.h"
#include "Log.h"
#include "TexLoad.h"


XList<Texture> TextureList;

GLint MaxTextureSize = 0;

static TextureLoader *Loader;



static int CheckTextureDimensions(GLsizei Width, GLsizei Height, GLint* numLODs);
static int FormatToBytesPerPixel(GLenum F);



int TexturesInit(void)
{
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MaxTextureSize);

	Loader = new(nothrow) TextureLoader;
	if (Loader == NULL) return 0;

	return 1;
}


void TexturesDone(void)
{
	delete Loader;
}


Texture* TextureLoadFromFile(char* S)
{
	Texture* T = new Texture;
	if (T != NULL)
	{
		int r = T->LoadTGA(S);
		if (!r)
		{
			delete T;
			return NULL;
		}
		TextureList.Insert(T);
	}
	return T;
}


Texture* TextureLoadFromMemory(void *P, GLsizei Width, GLsizei Height, GLenum Format)
{
	Texture* T = new Texture;
	if (T != NULL)
	{
		int r = T->LoadFromMemory(P, Width, Height, Format);
		if (!r)
		{
			delete T;
			return NULL;
		}
		TextureList.Insert(T);
	}
	return T;
}


void TextureDelete(Texture* T)
{
	if (T != NULL)
	{
		TextureList.Delete(T);
		delete T;
		T = NULL;
	}
}





int Texture::LoadFromMemory(void *P, GLsizei Width, GLsizei Height, GLenum Format)
{
	Unload();

	logInfo(modTextures, "Loading texture from memory (texture index %i).", Name);

	// load mode
	Mode = tlmMemory;

	// check texture dimensions and get LODsCount
	if (!CheckTextureDimensions(Width, Height, &LODsCount))
		{}; // ignore invalid texture dimensions

	// create TextureLODs
	LODs = new(nothrow) TextureLOD[LODsCount];
	if (LODs == NULL)
	{ logError(modTextures, "Out of memory."); return 0; }

	// get internal format (IF) from Format
	GLenum IF;
	switch (Format)
	{
		case GL_RGB  : IF = GL_RGB8; break;
		case GL_RGBA : IF = GL_RGBA8; break;
		case GL_LUMINANCE_ALPHA : IF = GL_LUMINANCE8_ALPHA8; break;
		default : IF = 1;
	}

	// initialize TextureLODs
	for (int i=0; i<LODsCount; ++i)
	{
		LODs[i].Initialize(this, i, IF, Width, Height, 0, Format);
		Width = Width >> 1;
		if (Width == 0) Width = 1;
		Height = Height >> 1;
		if (Height == 0) Height = 1;
	}

	// load data
	LODs[0].Data = (unsigned char*)P;

	// create next TextureLODs from previous one
	for (int i=1; i<LODsCount; ++i)
		LODs[i].MakeMemoryResident();

	// set texture parameters
	glBindTexture(GL_TEXTURE_2D, TexName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// send all TextureLODs to OpenGL
	for (int i=0; i<LODsCount; ++i)
		LODs[i].MakeGLResident();

	return 1;
}


int Texture::LoadFromFile(const char *AFileName)
{
	return 0;
}


int Texture::LoadTGA(char *AFileName)
{
	int i;

	Unload();

	logInfo(modTextures, "Loading texture \"%s\" through TGA loader (texture index %i).", AFileName, Name);

	// check FileName length
	i = strlen(AFileName);
	if (i>=MAX_TEXTURE_NAME_SIZE)
	{
		logError(modTextures, "Texture filename \"%s\" is too long, can't load texture.", AFileName);
		return 0;
	}
	strcpy(FileName, AFileName);

	// load mode
	Mode = tlmTGALoad;

	// ask loader to open file
	i = Loader->Open(AFileName);
	if (!i) return 0;

	// ask loader to read file header
	if (!Loader->IsHeaderLoaded())
	{ Loader->Close(); return 0; }

	// check size of texture, must be power of two
	unsigned int Size,X;
	unsigned int P;
	int L;
	if (Loader->TexWidth()>=Loader->TexHeight())
	{
		Size = Loader->TexWidth();
		X = Loader->TexHeight();
	} else {
		Size = Loader->TexHeight();
		X = Loader->TexWidth();
	}

	L = 1;
	P = 1;
	while (P<Size)
	{
		P *= 2;
		++L;
	}
	if (P != Size)
	{
		logError(modTextures, "Size of texture must be power of two (%s).", AFileName);
		return 0;
	}

	if (X != P)
	{
		P = 1;
		while (P<X)
		{
			P *= 2;
		}
		if (P != X)
		{
			logError(modTextures, "Size of texture must be power of two (%s).", AFileName);
			return 0;
		}
	}

	// create TextureLODs
	LODsCount = L;
	LODs = new(nothrow) TextureLOD[L];
	if (LODs == NULL)
	{ logError(modTextures, "Out of memory."); return 0; }

	// initialize TextureLODs
	GLsizei W = Loader->TexWidth();
	GLsizei H = Loader->TexHeight();
	GLenum F = Loader->TexFormat();
	GLint IF;
	switch (F)
	{
		case GL_RGB  : IF = GL_RGB8; break;
		case GL_RGBA : IF = GL_RGBA8; break;
		case GL_LUMINANCE_ALPHA : IF = GL_LUMINANCE8_ALPHA8; break;
		default : IF = 1;
	}

	for (i=0; i<L; ++i)
	{
		LODs[i].Initialize(this, i, IF, W, H, 0, F);
		W = W >> 1;
		if (W == 0) W = 1;
		H = H >> 1;
		if (H == 0) H = 1;
	}

	// read first TextureLOD from file to memory
	LODs[0].MakeMemoryResident();

	// create next TextureLODs from previous one
	for (i=1; i<LODsCount; ++i)
		LODs[i].MakeMemoryResident();

	// set texture parameters
	glBindTexture(GL_TEXTURE_2D, TexName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// send all TextureLODs to OpenGL
	for (i=0; i<LODsCount; ++i)
		LODs[i].MakeGLResident();

	return 1;
}


void Texture::Unload(void)
{
	Mode = tlmUnknown;
	delete[] LODs;
}



int TextureLOD::IntMakeGLResident(void)
{
	/* TODO: chybi pripad, kdy Data jsou NULL */
	glTexImage2D(GL_TEXTURE_2D, LOD, InternalFormat, Width, Height, Border, Format, GL_UNSIGNED_BYTE, Data);
	return 1;
}


int TextureLOD::MakeGLResident(void)
{
	glBindTexture(GL_TEXTURE_2D, Tex->TexName());
	/* TODO: chybi pripad, kdy Data jsou NULL */
	return IntMakeGLResident();
}


int TextureLOD::MakeMemoryResident(void)
{
	/* TODO: chybi pripad, kdy uz jsou Data nactena */
	int BPP = FormatToBytesPerPixel(Format);
	unsigned int Size = Width * Height * BPP;

	Data = (unsigned char*)malloc(Size);
	if (Data == NULL)
		return 0;

	if (LOD == 0)
	{
		switch (Tex->TexLoadMode())
		{
			case tlmTGALoad :
				// load file
				if (!Loader->StartTextureReading(Data, Size))
				{ Loader->Close(); return 0; }
				if (!Loader->IsTextureLoaded())
				{ Loader->Close(); return 0; }
				Loader->Close();
				break;

			case tlmFileLoad :
				break;

			case tlmMemory:
				break;

		}
	}
	else
//		if (1)
		{
			// reduction from LOD-1

			TextureLOD *PLOD = &Tex->TexLODs()[LOD-1];
			int LineSize = PLOD->Width * BPP;
			unsigned char *XData = PLOD->Data;

			unsigned char *CA = Data;
			unsigned char *CD = Data + Size;
			unsigned char *XA = XData;
			unsigned char *XB = XData + LineSize;
			unsigned char *XC = XB;

			switch (Format)
			{
				case GL_RGB  :
					{
						while (CA<CD)
						{
							while (XA<XC)
							{
								*(CA+0) = ((int)*(XA+0) + (int)*(XA+3) +
										   (int)*(XB+0) + (int)*(XB+3) + 2) >> 2;
								*(CA+1) = ((int)*(XA+1) + (int)*(XA+4) +
										   (int)*(XB+1) + (int)*(XB+4) + 2) >> 2;
								*(CA+2) = ((int)*(XA+2) + (int)*(XA+5) +
										   (int)*(XB+2) + (int)*(XB+5) + 2) >> 2;
								CA += 3;
								XA += 6;
								XB += 6;
							}
							XA = XB;
							XB += LineSize;
							XC = XB;
						}
					}
					break;
				case GL_RGBA :
					{
						while (CA<CD)
						{
							while (XA<XC)
							{
								*(CA+0) = ((int)*(XA+0) + (int)*(XA+4) +
										   (int)*(XB+0) + (int)*(XB+4) + 2) >> 2;
								*(CA+1) = ((int)*(XA+1) + (int)*(XA+5) +
										   (int)*(XB+1) + (int)*(XB+5) + 2) >> 2;
								*(CA+2) = ((int)*(XA+2) + (int)*(XA+6) +
										   (int)*(XB+2) + (int)*(XB+6) + 2) >> 2;
								*(CA+3) = ((int)*(XA+3) + (int)*(XA+7) +
										   (int)*(XB+3) + (int)*(XB+7) + 2) >> 2;
								CA += 4;
								XA += 8;
								XB += 8;
							}
							XA = XB;
							XB += LineSize;
							XC = XB;
						}
						break;
					}

/*		case GL_LUMINANCE_ALPHA :
			{
			while (CA<CD)
			{
			  Data[CA+0] = ((int)XData[XA+0] + (int)XData[XA+2] +
							(int)XData[XB+0] + (int)XData[XB+2] + 2) >> 2;
			  Data[CA+1] = ((int)XData[XA+1] + (int)XData[XA+3] +
							(int)XData[XB+1] + (int)XData[XB+3] + 2) >> 2;
			  CA += 2;
			  XA += 4;
			  XB +- 4;
			}
		  }
		  break;
		default :
			{
			while (CA<CD)
			{
			  Data[CA+0] = ((int)XData[XA+0] + (int)XData[XA+1] +
							(int)XData[XB+0] + (int)XData[XB+1] + 2) >> 2;
			  CA += 1;
			  XA += 2;
			  XB +- 2;
			}
		  }*/
		}
	}
	return 1;
}


static int FormatToBytesPerPixel(GLenum F)
{
  switch (F)
  {
	case GL_RGB  : return 3;
	case GL_RGBA : return 4;
	case GL_LUMINANCE_ALPHA : return 2;
	default : return 1;
  }
}



static int CheckTextureDimensions(GLsizei Width, GLsizei Height, GLint* numLODs)
{
	unsigned int Size,X;
	unsigned int P;
	int L;

	// assign longer dimension of texture to Size and shorter to X
	if (Width>=Height)
	{
		Size = Width;
		X = Height;
	} else {
		Size = Height;
		X = Width;
	}

	// check longer dimension
	L = 1;
	P = 1;
	while (P<Size)
	{
		P *= 2;
		++L;
	}
	if (P != Size)
	{
		logError(modTextures, "Size of texture must be power of two.");
		*numLODs = L;
		return 0;
	}

	// check shorter dimension
	if (X != P)
	{
		P = 1;
		while (P<X)
		{
			P *= 2;
		}
		if (P != X)
		{
			logError(modTextures, "Size of texture must be power of two.");
			*numLODs = L;
			return 0;
		}
	}

	*numLODs = L;
	return 1;

}

